from argparse import ArgumentParser
from collections import defaultdict
from datetime import datetime
from io import DEFAULT_BUFFER_SIZE, StringIO
import subprocess
import os
import sys
from typing import Dict, IO, Optional, Sequence, Tuple
import warnings

import opa_tools # type: ignore (configure warnings)

MSG_KEYS = ("Tagger", "Date", "Merge", "Author", "commit", "tag")
DATE_FORMAT_IN = r"%a %b %d %H:%M:%S %Y %z"
DATE_FORMAT_OUT = r"%Y/%m/%d"

TYPE_WORDS = {
    "added": ("add",),
    "changed": ("change",),
    "fixed": ("fix", "bug"),
    "deprecated": (),
    "removed": (),
    "security": (),
}
DEFAULT_TYPE_WORD = "changed"

def _get_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        type=str,
        help="Directory of the package whose changelog is to be written"
    )
    return parser

def process_tag_info(annotation, stopper="commit") -> Tuple[str, Dict[str, str]]:
    info = {}
    io = StringIO()
    for line in annotation.split("\n"):
        line = line.strip(" \n")
        if (
            stopper
            and line.startswith(stopper)
            and "date" in info
        ):
            break
        for key in MSG_KEYS:
            if line.strip().startswith(key):
                info[key.lower()] = (
                    line
                    .removeprefix(key)
                    .removeprefix(key.lower())
                    .strip(": ")
                )
                break
        else:
            if not line.strip():
                continue
            io.write(line.strip() + "\n")
    return io.getvalue(), info

def get_git_tags():
    try:
        # Run the git tag command to get all the tags
        result = subprocess.run(["git", "tag"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        
        if result.returncode != 0:
            print(f"Error getting tags: {result.stderr}")
            return []
        
        # Split the output into a list of tags
        tags = result.stdout.strip().split('\n')
        
        return tags if tags != [''] else []  # Return empty list if no tags are found
    except Exception as e:
        print(f"An error occurred: {e}")
        return []

def get_git_tags_with_annotations():
    try:
        # Get all tags
        result = subprocess.run(
            ["git", "tag"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        if result.returncode != 0:
            print(f"Error getting tags: {result.stderr}")
            return []
        tags = result.stdout.strip().split('\n')
        tag_details = []
        for tag in tags:
            if tag:  # If the tag is not an empty string
                # Get the annotation for the tag using 'git show'
                show_result = subprocess.run(
                    ["git", "show", "-s", tag],
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
                )
                if show_result.returncode != 0:
                    print(f"Error getting details for tag {tag}: {show_result.stderr}")
                    continue
                tag_info = show_result.stdout.strip()
                tag_desc, tag_info = process_tag_info(tag_info)
                tag_details.append((tag, tag_desc, tag_info))
        return tag_details
    except Exception as e:
        print(f"An error occurred: {e}")
        return []

def parse_tag_date(date_str: str, format_out: Optional[str] = None):
    if format_out is None:
        format_out = DATE_FORMAT_OUT
    date = datetime.strptime(date_str, DATE_FORMAT_IN)
    return date.strftime(format_out)

def sanitize_markdown(text):
    return text.replace("#", r"\#")

def _get_other_words(exclude_type):
    other_words = []
    for type_, words in TYPE_WORDS.items():
        if type_ == exclude_type:
            continue
        other_words.extend(other_words)
    return tuple(sorted(other_words))

def classify_type(desc, default=""):
    if default and not default in TYPE_WORDS:
        raise ValueError(
            f"Default value '{default}' is not a type word."
        )
    desc = desc.lower()
    for type_, words in TYPE_WORDS.items():
        other_words = _get_other_words(type_)
        if (
            any(word in desc for word in words)
            and not any(word in desc for word in other_words)
        ):
            return type_.capitalize()
    return default.capitalize()

def process_description(description):
    return tuple(
        (sanitize_markdown(line), classify_type(line, DEFAULT_TYPE_WORD))
        for line in description.strip().split("\n")
    )


def process_git_tags(tags):
    filtered_tags = list(filter(
        lambda x: x[0].startswith("v"),
        tags
    ))
    if len(filtered_tags) < len(tags):
        difference = (
            set(el[0] for el in tags)
            - set(el[0] for el in filtered_tags)
        )
        warnings.warn(
            "Dropped tags:\n- "
            "\n- ".join(difference)
        )
    #if "date" not in info:
    #    warnings.warn(
    #        f"No info in tag {name}."
    #    )
    name_date_topics = [
        (
            name,
            parse_tag_date(info["date"]),
            process_description(desc),
        )
        for name, desc, info in filtered_tags
    ]
    sorted_name_date_topics = sorted(
        (tup3 for tup3 in name_date_topics),
        key=lambda tup4: (tup4[1], tup4[0]),
        reverse=True,
    )
    return sorted_name_date_topics

def _line_cleanup(line: str) -> str:
    return (
        line
        .strip()
        .removeprefix("- ")
        .capitalize()
    )

def write_section(topics: Sequence[Tuple[str, str]]) -> str:
    subsections = StringIO()
    subsections_dct = defaultdict(list)
    type_words = tuple(TYPE_WORDS.keys())
    for desc, type_ in topics:
        subsections_dct[type_].append(desc)
    sorted_subsections = sorted(
        list(subsections_dct.items()),
        key=lambda x: type_words.index(x[0].lower()) if x[0] else -1
    )

    for type_, lines in sorted_subsections:
        subsections.write(f"### {type_}\n")
        subsections.write("- " + "\n- ".join(
            _line_cleanup(line) for line in lines
        ))
        subsections.write("\n\n")
    return subsections.getvalue()

def write_changelog(name_date_topics, buffer: IO):
    buffer.write("# Changelog\n\n")
    for name, date, topics in name_date_topics:
        section_header = f"## {name} - {date}"
        section = write_section(topics)
        buffer.write(
            section_header + "\n" + section + "\n"
        )

if __name__ == "__main__":
    args = _get_parser().parse_args(sys.argv[1:])
    os.chdir(args.path)
    tags_info_raw = get_git_tags_with_annotations()
    tags_info = process_git_tags(tags_info_raw)
    changelog = StringIO()
    write_changelog(tags_info, changelog)
    with open("CHANGELOG.md", "w") as wf:
        wf.write(changelog.getvalue())

